// BackpropNetwork.cc
// Dan Pemstein
// Little Celtic Warlock Software 2001
// ANN++ Library 0.1

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.

#ifndef BACKPROP_NETWORK_CC
#define BACKPROP_NETWORK_CC

#include "BackpropNetwork.h"

BackpropNetwork::BackpropNetwork(size_t layers, size_t *nodes_per)
	: Network (layers, nodes_per),
		bias_ (new BackpropNeuron(new InputUnit()))
{
}

BackpropNetwork::BackpropNetwork (const char *fname)
	:	Network(fname),
		bias_ (new BackpropNeuron(new InputUnit()))
{
	ifstream inFile(fname);
	if ( ! inFile ) {
		cerr << "Cannot open " << fname << " for reading.\n"
			<< "Aborting BackpropNetwork::BackpropNetwork(const char *)!\n";
		exit(-1);
	}
	size_t pos[4];
	double weight;
	WeightedConnection *con;
	inFile.ignore(20,'\n');
	while ( ! inFile.eof() ) {
		inFile >> pos[0] >> pos[1] >> pos[2] >> pos[3];
		if (inFile.eof())
			break;
		inFile >> weight;
		if (pos[0] == in_) {
			addNeuron(pos[0], pos[1], returnInputNeuron());
		} else if (pos[0] >= num_layers_) {
			if ( (! finalized_) || pos[0] > num_layers_) {
				cerr << "Out-of-range node in infile.\n"
				<< "Aborting BackpropNetwork::BackpropNetwork(const char *)!\n";
				exit(-1);
			}
			break;
		} else {
			addNeuron(pos[0], pos[1], returnNeuron());
		}
		if (pos[2] == in_) {
			addNeuron(pos[2], pos[3], returnInputNeuron());
		} else {
			addNeuron(pos[2], pos[3], returnNeuron());
		}
		con = (WeightedConnection*) returnConnection();
		con->weight(weight);
		connect(pos[0], pos[1], pos[2], pos[3], con);
	}
	if (finalized_) {
		Neuron *n;
		con = (WeightedConnection*) returnConnection();
		con->weight(weight);
		// We have to get a little closer to the metal here caus
		// connect() won't work with the bias unit
		layers_[pos[2]].get(pos[3], n);
		con->value(n);
		con->position(num_layers_, num_layers_, pos[2], pos[3]);
		n->addInput();
		bias_->addOutput(con);
		while (! inFile.eof() ) {
			inFile >> pos[0] >> pos[1] >> pos[2] >> pos[3];
			if (inFile.eof())
				break;
			inFile >> weight;
			con = (WeightedConnection*) returnConnection();
			con->weight(weight);
			// We have to get a little closer to the metal here caus
			// connect() won't work with the bias unit
			layers_[pos[2]].get(pos[3], n);
			con->value(n);
			con->position(num_layers_, num_layers_, pos[2], pos[3]);
			n->addInput();
			bias_->addOutput(con);
		}
		bias_->addInput();
		Network::finalize();
	}
	inFile.close();
}

BackpropNetwork::~BackpropNetwork()
{
	delete bias_;
}

void
BackpropNetwork::finalize()
{
	// Add and connect the bias unit.  We have to do this by hand 
	// because the bias
	// unit is not within any network layer and is invisible to the
	// base network class.
	size_t j;
	double r;
	Neuron *n;
	Connection *con;
	LayerIterator *li = new LayerIterator(&(layers_[in_]));
	for (size_t i = 1; i <= out_; i++) { // for every layer
		li->iterateOver(&(layers_[i]));
		while (! li->is_done()) {
			li->current_item(j);
			layers_[i].get(j, n);
			r = (drand48() * .1) - .05;
			//r = (drand48() * 2) - 1;
			con = new WeightedConnection(n,0,r);
			con->position(num_layers_, num_layers_, i, j);
			n->addInput();
			bias_->addOutput(con);
			li->next();
		}
	}
	bias_->addInput(); // don't forget this
	// Do the rest of finalize
	Network::finalize();
}

void
BackpropNetwork::run(double *invec, double *outvec)
{
	// Charge the bias unit
	bias_->placeCharge(1);
	Network::run(invec, outvec);
}

void 
BackpropNetwork::train(const size_t &n, double *invec, 
                       const double &rate)
{
	double t, o, e;
	size_t index = 0;
	size_t temp_index;
	Neuron *temp_node;
	BackpropNeuron *temp_node2;
	Connection *temp_conn;
	LayerIterator *li = new LayerIterator(&(layers_[in_]));
	
	for (size_t i = 0; i < n; i++) { // example iterations
		// Charge the bias
		bias_->placeCharge(1);
		// Feed input into network
		li->iterateOver(&(layers_[in_]));  // redundant first time
		while (! li->is_done()) {
			li->current_item(temp_index);
			layers_[in_].get(temp_index, temp_node);
			(*temp_node).placeCharge(invec[index]);
			index++;
			li->next();
		}
		// Get the output values and calculate errors of output units
		temp_conn = output_;
		li->iterateOver(&(layers_[out_]));
		for (; ! li->is_done(); index++) {
			t = invec[index];
			o = temp_conn->getCharge();
			e = o * (1 - o) * (t - o);
			li->current_item(temp_index);
			layers_[out_].get(temp_index, temp_node);
			temp_node2 = static_cast<BackpropNeuron*> (temp_node);
			(*temp_node2).setError(e);
			temp_conn = temp_conn->next();
			li->next();
		}
		// Now for the hidden layers and weight updates
		for (size_t j = out_ - 1; j >= in_ && j < out_; j--) {// each layer
			li->iterateOver(&(layers_[j]));
			while (! li->is_done()) { // each neuron in the layer
				li->current_item(temp_index);
				layers_[j].get(temp_index, temp_node);
				temp_node2 = static_cast<BackpropNeuron*> (temp_node);
				temp_node2->calcError(rate, n);
				li->next();
			}
		}
		// train the bias weights
		bias_->calcError(rate, n);
	}	
	delete li;
}

void 
BackpropNetwork::save(const char *fname)
{
	ofstream outFile(fname);
	if ( ! outFile ) {
		cerr << "Cannot open " << fname << " for writing.\n"
				 << "Aborting BackpropNetwork::save(const char*)!\n";
		return;
	}
	// Write a header containing the number of layers and the nodes per
	outFile << num_layers_;
	for ( size_t i = 0; i < num_layers_; i++) {
		outFile << " " << layers_[i].size();
	}
	// Continue on with in_ and out_;
	outFile << " " << in_ << " " << out_;
	// And a finalized flag
	if ( finalized_ )
		outFile << " " << 1;
	else
		outFile << " " << 0;
	
	// Now we build a connection table
	size_t tempIndex;
	Neuron *tempNode;
	Connection *tempConn;
	size_t pos[4] = {0, 0, 0, 0};
	WeightedConnection *tempConn2;
	layers_[0].get(0, tempNode);  // we need a dummy to create the ni
	LayerIterator *li = new LayerIterator(&(layers_[in_]));
	NeuronIterator *ni = new NeuronIterator(tempNode);
	for ( size_t i = 0; i < num_layers_; i++ ) {
		if (i != out_) {
			li->iterateOver(&(layers_[i]));
			while ( ! li->is_done()) {
				li->current_item(tempIndex);
				layers_[i].get(tempIndex, tempNode);
				ni->iterateOver(tempNode);
				while ( ! ni->is_done()) {
					ni->current_item(tempConn);
					tempConn2 = static_cast<WeightedConnection*> (tempConn);
					tempConn2->position(pos);
					outFile << endl;
					outFile << pos[0] << " " << pos[1] << " " 
									<< pos[2] << " " << pos[3] << " ";
					outFile << setprecision(20) << tempConn2->weight();
					ni->next();
				}
				li->next();
			}
		}
	}
	if ( finalized_ ) { // save bias unit info if necessary
		tempNode = bias_;
		ni->iterateOver(tempNode);
		while ( ! ni->is_done()) {
			ni->current_item(tempConn);
			tempConn2 = static_cast<WeightedConnection*> (tempConn);
			tempConn2->position(pos);
			outFile << endl;
			outFile << pos[0] << " " << pos[1] << " " << pos[2] << " "
			        << pos[3] << " ";
			outFile << setprecision(20) << tempConn2->weight();
			ni->next();
		}
	}
	outFile.close();
	delete ni;
	delete li;
}

Neuron
*BackpropNetwork::returnInputNeuron()
{
	Neuron *n = new BackpropNeuron(new InputUnit());
	return n;
}

Neuron
*BackpropNetwork::returnNeuron()
{
	Neuron *n = new BackpropNeuron(new SigmoidUnit());
	return n;
}

Connection
*BackpropNetwork::returnConnection()
{
	double r = (drand48() * .1) - .05;
	//double r = (drand48() * 2) - 1;
	Connection *c = new WeightedConnection(0,0,r);
	return c;
}

#endif /* BACKPROP_NETWORK_CC */
