// Network.cc
// Dan Pemstein
// Little Celtic Warlock Software 2001
// ANN++ Library 0.1

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.

#ifndef NETWORK_CC
#define NETWORK_CC

#include "Network.h"

Network::Network (size_t layers, size_t *nodes_per)
	: layers_ (0),
		in_ (0),
		out_ (layers - 1),
		num_layers_ (layers),
		output_ (0),
		finalized_ (0)
{
	layers_ = new (std::nothrow) Layer[num_layers_];
	if (layers_ == 0) {
		cerr << "Network::Network(size_t, size_t *); allocation of "
		     << layers << " layers failed!" << endl;
		exit (-1);
	}
	for (size_t i = 0; i < num_layers_; i++) {
		layers_[i].resize(nodes_per[i]);
	}
}

Network::Network (const char *fname)
	:	layers_ (0),
		in_ (0),
		out_ (0),
		num_layers_ (0),
		output_ (0),
		finalized_ (0)
{
	// This only takes care of basic setup of an empty network
	// Extended class constructors must actually fill net, etc.
	ifstream inFile(fname);
	if ( ! inFile ) {
		cerr << "Cannot open " << fname << " for reading.\n"
		     << "Aborting Network::Network(const char*)!\n";
		exit (-1);
	}
	inFile >> num_layers_;
	layers_ = new (std::nothrow) Layer[num_layers_];
	if (layers_ == 0) {
		cerr << "Network::Network(size_t, size_t *); allocation of "
		     << num_layers_ << " layers failed!" << endl;
		exit (-1);
	}
	size_t per[num_layers_];
	for ( size_t i = 0; i < num_layers_; i++ ) {
		inFile >> per[i];
		layers_[i].resize(per[i]);
	}
	inFile >> in_;
	inFile >> out_;
	inFile >> finalized_;
	inFile.close();
}

Network::~Network ()
{
	// This removes and cleans up all the output connections
	Connection *temp_head;
	while (output_ != 0) {
		temp_head = output_->next();
		delete output_;
		output_ = temp_head;
	}

	// Deleting layers will remove all neurons and do dynamic mem
	// cleanup
	delete[] layers_;
}

int
Network::setOutput(const size_t &i) {
	if (i < num_layers_) {
		out_ = i;
		return 0;
	}
	return -1;
}

void
Network::fill (Neuron * (*in)(), Neuron * (*rn)())
{
	Neuron * (*funcpoint)() = in;
	fill(in_, funcpoint);
	Neuron * (*funcpoint2)() = rn;
	for (size_t i = 1; i < num_layers_; i++)
		fill(i, funcpoint2);
}

int
Network::fill (size_t & layer, Neuron * (*rn)())
{
	if (layer >= num_layers_)
		return -1;
	size_t i;
	LayerIterator *li = 
		new (std::nothrow) LayerIterator(&(layers_[layer]));
	li->first();
	while (! li->is_done()) {
		li->current_item(i);
		addNeuron(layer, i, rn());
		li->next();
	}
	delete li;
	return 0;
}

void
Network::empty ()
{
	for (size_t i = 1; i < num_layers_; i++)
		empty(i);
}

void
Network::empty(const size_t &layer)
{
	layers_[layer].empty();
}

void
Network::finalize ()
{
	size_t cur;
	Neuron *node;
	Connection *con;
	Connection *prev;
	bool first = 1;
	LayerIterator *li = 
		new (std::nothrow) LayerIterator(&(layers_[out_]));
	li->first();
	while (! li->is_done()) {
		li->current_item(cur);
		layers_[out_].get(cur, node);
		con = new (std::nothrow) Connection();
		(*node).addOutput(con);
		
		if (first) {
			output_ = con;
			first = 0;
		} else {
			prev->next(con);
		}
		prev = con;
		li->next();
	}
	li->iterateOver(&(layers_[in_]));
	li->first();
	while (! li->is_done()) {
		li->current_item(cur);
		layers_[in_].get(cur, node);
		(*node).addInput();
		li->next();
	}
	delete li;
	finalized_ = 1;
}

int
Network::addNeuron(const size_t &layer, const size_t &index,
                   Neuron * node)
{
	if (layer < num_layers_) 
		if (layers_[layer].addNeuron(node, index)) 
			return 0;
	return -1;
}

int
Network::removeNeuron(const size_t &layer, const size_t &index)
{
	if (layer < num_layers_)
		if (layers_[layer].removeNeuron(index))
			return 0;
	return -1;
}

void
Network::connect(Connection * (*rc) ())
{
	Connection * (*funcpoint)() = rc;
	for (size_t i = 0; i < (num_layers_ - 1); i++)
		connect(i, i+1, funcpoint);
}

void
Network::disconnect()
{
	for (size_t i = 0; i < (num_layers_ - 1); i++)
		disconnect(i, i+1);
}

int
Network::connect(const size_t &a, const size_t &b,
                 Connection * (*rc)())
{
	Connection * (*funcpoint)() = rc;
	if (a >= num_layers_ || b >= num_layers_)
		return -1;
	LayerIterator *lia =
		new (std::nothrow) LayerIterator(&(layers_[a]));
	LayerIterator *lib =
		new (std::nothrow) LayerIterator(&(layers_[b]));
	size_t i;
	size_t j;
	lia->first();
	while (! lia->is_done()) {
		lib->first();
		lia->current_item(i);
		while (! lib->is_done()) {
			lib->current_item(j);
			connect(a, i, b, j, funcpoint());
			lib->next();
		}
		lia->next();
	}
	delete lia;
	delete lib;
}

int
Network::disconnect(const size_t &a, const size_t &b)
{
	if (a >= num_layers_ || b >= num_layers_)
		return -1;
	LayerIterator *lia =
		new (std::nothrow) LayerIterator(&(layers_[a]));
	LayerIterator *lib =
		new (std::nothrow) LayerIterator(&(layers_[b]));
	size_t i;
	size_t j;
	lia->first();
	while (! lia->is_done()) {
		lib->first();
		lib->current_item(i);
		while (! lib->is_done()) {
			lib->current_item(j);
			disconnect(a, i, b, j);
			lib->next();
		}
		lia->next();
	}
	delete lia;
	delete lib;
}

int
Network::connect(const size_t &la, const size_t &na,
                 const size_t &lb, const size_t &nb,
								 Connection *con)
{
	if (la >= num_layers_ || lb >= num_layers_)
		return -1;
	Neuron *nea;
	Neuron *neb;
	if (layers_[la].get(na, nea) != 0)
		return -2;
	if (layers_[lb].get(nb, neb) != 0)
		return -3;
	(*con).value(neb);
	(*con).position(la, na, lb, nb);
	if ((*neb).addInput() != 0)
		return -4;
	if ((*nea).addOutput(con) != 0)
		return -5;
	return 0;
}

int
Network::disconnect(const size_t &la, const size_t &na,
                    const size_t &lb, const size_t &nb)
{
	if (la >= num_layers_ || lb >= num_layers_)
		return -1;
	Neuron *nea;
	Neuron *neb;
	if (layers_[la].get(na, nea) != 0)
		return -2;
	if (layers_[lb].get(nb, neb) != 0)
		return -3;
	if ((*neb).removeInput() != 0)
		return -4;
	if ((*nea).removeOutput(*neb) != 0)
		return -5;
}	

void
Network::run(double *invec, double *outvec)
{
	Neuron *temp;
	size_t cur;
	size_t j = 0;
	LayerIterator *li =
		new (std::nothrow) LayerIterator(&(layers_[in_]));
	li->first();
	while (! li->is_done()) {
		li->current_item(cur);
		layers_[in_].get(cur, temp);
		(*temp).placeCharge(invec[j]);
		j++;
		li->next();
	}
	delete li;
	Connection *temp2 = output_;
	j = layers_[out_].size();
	for (size_t i = 0; i < j; i++) {
		outvec[i] = (*temp2).getCharge();
		temp2 = (*temp2).next();
	}
}

#endif /* NETWORK_CC */
