// Network.h
// Dan Pemstein
// Little Celtic Warlock Software 2001
// ANN++ Library 0.1

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.

#ifndef NETWORK_H
#define NETWORK_H

#include <new>
#include <stdlib.h>
#include <fstream.h>
#include "Neuron.h"
#include "Connection.h"
#include "Layer.h"

class Neuron;	// forward declaration

class Network
{
public:
	// Destructor
	virtual ~Network ();

	// Set the index of the output layer
	int setOutput(const size_t &);

	// Fill the network with nodes.  The first function pointer is to
	// a func that generates the appropriate input nodes for this network
	// the second is for other nodes.  All functions taking pointers
	// typically expect to be fed functions from extending classes, but
	// users may use any legal function if they wish for flexibility.
	void fill(Neuron * (*)(), Neuron * (*)());

	// Fill a layer with nodes.  Same note as above.
	int fill(size_t &, Neuron * (*)());

	// Empty the network.  (Also deletes objects freeing the user of
	// the need (a must if fill() is used)).
	void empty();

	// Empty a layer
	void empty (const size_t &);

	// Do whatever is necessary (such as attaching output connections)
	// to make this network run.  Subclasses may override this method
	// for added features.  Must be called before run or train.
	virtual void finalize();

	// Add a Neuron to the specified layer and index.
	int addNeuron(const size_t &, const size_t &, Neuron *);

	// Remove a particular Neuron from the specified layer.
	int removeNeuron(const size_t &, const size_t &);

	// Fully connect the network.  
	void connect(Connection * (*)());

	// Fully disconnect network
	void disconnect();

	// Fully connect two layers.  Same note as above.
	int connect(const size_t &, const size_t &, Connection * (*)());

	// Fully disconnect two layers
	int disconnect(const size_t &, const size_t &);

	// Connect two nodes (nodes are indexed within layers from
	// "left" to "right").  Will connect them again if they are
	// already connected.
	int connect(const size_t &, const size_t &,
	            const size_t &, const size_t &,
			        Connection *);
	
	// Disconnect two nodes (nodes are indexed within layers from
	// "left" to "right")
	int disconnect(const size_t &, const size_t &,
	               const size_t &, const size_t &);

	// Train the network for n cycles.  
	// Pass vector (array) of inputs and 
	// correct outputs in order (i.e xor = {0,0,0, 1,0,1,
	// 0,1,1, 1,1,0}).  The network infers the input and output
	// size from the sizes of the input and output layers.
	virtual void train(const size_t &, double *, const double &) = 0;
	
	// Run the network on an example without any training update
	// Pass in a vector representing one example.
	// The network infers the number of
	// elements per example from the size of the input layer.
	// Returns the output vector to the second doulbe
	virtual void run(double *, double *);

	// Save the network to file for later use.  This is a completely
	// algorithm-specific function and must be implemented by
	// extending classes
	virtual void save(const char*) = 0;

protected:
	// Create a network with the specified number of layers and
	// the max number of neurons in each layer.
	Network (size_t, size_t *);

	// A constructor that loads a network from file
	Network (const char *);

	// Array of layer containers
	Layer *layers_;

	// Index of the input layer (default 0)
	size_t in_;

	// Reference of the output layer (default num_layers_ - 1)
	size_t out_;

	// Number of layers in the network
	size_t num_layers_;

	// The linked-list of output connections
	Connection *output_;

	// Finalized (inputs and outputs attached/cleanup) flag
	bool finalized_;
};

#endif /* NETWORK_H */
