// Neuron.h
// Dan Pemstein
// Little Celtic Warlock Software 2001-2002
// ANN++ Library 0.1

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.

#ifndef NEURON_H
#define NEURON_H

#include <new>
#include <stdlib.h>  // for size_t type
#include "Connection.h"
#include "Unit.h"

class Connection;		// forward declaration
class NeuronIterator;

class Neuron
{
// Somewhat "abstract" neuron class which takes care of general
// functionality but leaves implementation of learning rules to
// derived classes.  All methods are virtual for easy overriding.
// (The overhead is really pretty minimal)
public:
	// Add an incoming neural connection (increment counter)
	virtual int addInput();
		
	// Add an outgoing neural connection
	virtual int addOutput(Connection *);

	// Remove an incoming connection (decrements counter)
	virtual int removeInput();

	// Remove an outgoing connection
	virtual int removeOutput(Neuron &);

	// Apply an incomming "charge" to the neuron.
	void placeCharge(const double &);

	// Destructor
	virtual ~Neuron();

	friend class NeuronIterator;

protected:
	// Constructor
	// Takes a connection list (or 0), the number of outs, ins,
	// excited (nomrally all 0) and its unit type.
	Neuron (Connection *, size_t, size_t, size_t, Unit *);

	// Linked list of output connections
	Connection *output_;
	
	// The number of output connections
	size_t num_outputs_;
	
	// The number of input connections
	size_t num_inputs_;
	
	// The number of input connections that have place input since
	// the last time this neuron fired.
	size_t excite_level_;

	// The Unit encapsulated by this Neuron.  Can be sigmoid, tanh
	// or other.  The unit does the actual calculations of the neuron
	Unit *unit_;

};

class NeuronIterator
{
public:
	// Constructor
	NeuronIterator (const Neuron *);

	// Default constructor
	NeuronIterator();

	// Destructor
	~NeuronIterator ();

	void iterateOver(const Neuron *n);

	// Iterator methods
	void first ();

	int next ();

	int is_done () const;

	int current_item (Connection *&) const;

private:
	// The neuron being iterated over
	const Neuron *ineuron_;

	// The current item's index
	Connection *current_;
};

#endif /* NEURON_H */
